/*****************************************************************************
 * Copyright (c) 2022, Nations Technologies Inc.
 *
 * All rights reserved.
 * ****************************************************************************
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer below.
 *
 * Nations' name may not be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * DISCLAIMER: THIS SOFTWARE IS PROVIDED BY NATIONS "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * DISCLAIMED. IN NO EVENT SHALL NATIONS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ****************************************************************************/

/**
 * @file FlashPrg.c
 * @author Nations
 * @version V1.1.0
 *
 * @copyright Copyright (c) 2022, Nations Technologies Inc. All rights reserved.
 */
 
#include "FlashOS.h"        // FlashOS Structures

typedef volatile unsigned char  vu8;
typedef volatile unsigned long  vu32;
typedef volatile unsigned short vu16;

#define M8(adr)  (*((vu8  *) (adr)))
#define M16(adr) (*((vu16 *) (adr)))
#define M32(adr) (*((vu32 *) (adr)))

// Peripheral Memory Map
#define IWDG_BASE       (0x40003000)
#define RCC_BASE        (0x40021000)
#define FLASH_BASE      (0x40022000)
#define PWR_BASE        (0x40007000)
#define SRAM_ADDR       (0x40001820)

#define IWDG            ((IWDG_TypeDef *) IWDG_BASE)
#define RCC             ((RCC_TypeDef *)  RCC_BASE)
#define FLASH           ((FLASH_TypeDef*) FLASH_BASE)
#define PWR             ((PWR_TypeDef*)   PWR_BASE)

// FLASH BANK size
#define BANK1_SIZE      0x00020000      // Bank1 Size = 128kB
// System Memory size
#define SYSMEMORY_SIZE  0x00004000      // System Memory Size = 16kB

// Independent WATCHDOG
typedef struct {
  vu32 KR;
  vu32 PR;
  vu32 RLR;
  vu32 SR;
} IWDG_TypeDef;

// Flash Registers
typedef struct {
  vu32 ACR;                                     // offset  0x000
  vu32 KEYR;                                    // offset  0x004
  vu32 OPTKEYR;                                 // offset  0x008
  vu32 SR;                                      // offset  0x00C
  vu32 CR;                                      // offset  0x010
  vu32 AR;                                      // offset  0x014
  vu32 OBR2;                                    // offset  0x018
  vu32 OBR;                                     // offset  0x01C
  vu32 WRPR;                                    // offset  0x020
} FLASH_TypeDef;

typedef struct
{
    vu32 CTRL1;
    vu32 CTRL2;
    vu32 CTRL3;
    vu32 STS1;
    vu32 STS2;
    vu32 STSCLR;
} PWR_TypeDef;

typedef struct
{
    vu32 CTRL;
    vu32 CFG;
    vu32 CLKINT;
    vu32 APB2PRST;
    vu32 APB1PRST;
    vu32 AHBPCLKEN;
    vu32 APB2PCLKEN;
    vu32 APB1PCLKEN;
    vu32 LDCTRL;
    vu32 CTRLSTS;
    vu32 AHBPRST;
    vu32 CFG2;
    vu32 CFG3;
    vu32 RDCTRL;
    vu32 Reserve0;
    vu32 Reserve1; 
    vu32 PLLHSIPRE;
    vu32 SRAM_CTRLSTS;
} RCC_TypeDef;

// Flash Keys
#define RDPRT_KEY       (0x5AA5)
#define FLASH_KEY1      (0x45670123)
#define FLASH_KEY2      (0xCDEF89AB)

// Flash Control Register definitions
#define FLASH_PG        (0x00000001)
#define FLASH_PER       (0x00000002)

#define CR_PER_Reset    (0x00001FFD)

#define FLASH_MER       (0x00000004)
#define FLASH_OPTPG     (0x00000010)
#define FLASH_OPTER     (0x00000020)
#define FLASH_STRT      (0x00000040)
#define FLASH_LOCK      (0x00000080)
#define FLASH_OPTWRE    (0x00000200)

// Flash Status Register definitions
#define FLASH_BSY       (0x00000001)
#define FLASH_PGERR     (0x00000004)
#define FLASH_WRPRTERR  (0x00000010)
#define FLASH_EOP       (0x00000020)

#define FLASH_PROG_MASK (FLASH_OPTWRE|FLASH_OPTER|FLASH_OPTPG|FLASH_MER|FLASH_PER|FLASH_PG)

#define PWR_CTRL1_MRSEL_0       (0x00000200)
#define PWR_CTRL1_MRSEL_1       (0x00000400)
#define PWR_CTRL1_MRSEL_MASK    (PWR_CTRL1_MRSEL_0|PWR_CTRL1_MRSEL_1)
#define PWR_STS2_MRF            (0x00000002)

// RCC register definitiions
#define RCC_HSI_ENABLE          (0x00000001)
#define RCC_CTRL_HSIRDF         (0x00000002)
#define RCC_PWR_CLOCK_ENABLE    (0x10000000)

//unsigned long base_adr;


/*
 *  Initialize Flash Programming Functions
 *    Parameter:      adr:  Device Base Address
 *                    clk:  Clock Frequency (Hz)
 *                    fnc:  Function Code (1 - Erase, 2 - Program, 3 - Verify)
 *    Return Value:   0 - OK,  1 - Failed
 */

#ifdef FLASH_MEM
int Init (unsigned long adr, unsigned long clk, unsigned long fnc) 
{
    (void)adr;
    (void)clk;
    (void)fnc;
  
    unsigned long temp;
    
    __disable_irq();
    
    IWDG->KR = 0xAAAA;
    
    // Test if IWDG is running (IWDG in HW mode)
//    if ((FLASH->OBR & 0x04) == 0x00) 
    {
        // Set IWDG time out to ~32.768 second
        IWDG->KR  = 0x5555;     // Enable write access to IWDG_PR and IWDG_RLR     
        IWDG->PR  = 0x06;       // Set prescaler to 256  
        IWDG->RLR = 4095;       // Set reload value to 4095
    }
    
    IWDG->KR = 0xAAAA;
    
    RCC->CTRL       |= RCC_HSI_ENABLE;          /*Enable HSI*/
    RCC->APB1PCLKEN |= RCC_PWR_CLOCK_ENABLE;    /*Enable PWR clock */
    FLASH->ACR      |= 0x00000003;  /*flash latency set to maximum value*/
    
    /* Wait MR to stablilize */
    temp = 65*25;
    while(temp > 0)
    {
        temp--;
        if((PWR->STS2 & PWR_STS2_MRF) == PWR_STS2_MRF)
            break;
    }

    /* Change MR to 1.1V */
    PWR->CTRL1 |= PWR_CTRL1_MRSEL_MASK;

    /* Wait MR to start changing */
    temp = 65*25;
    while(temp > 0)
    {
        temp--;
        if((PWR->STS2 & PWR_STS2_MRF) != PWR_STS2_MRF)
            break;
    }
    
    /* Wait MR to stablilize */
    temp = 65*25;
    while(temp > 0)
    {
        temp--;
        if((PWR->STS2 & PWR_STS2_MRF) == PWR_STS2_MRF)
            break;
    }
    
    M32(SRAM_ADDR) &= ~(0x1 <<25);
    
    __enable_irq();
    
    while((RCC->CTRL & RCC_CTRL_HSIRDF) != RCC_CTRL_HSIRDF);
    
    // Unlock Flash    
    FLASH->KEYR  = FLASH_KEY1;
    FLASH->KEYR  = FLASH_KEY2;

//    base_adr = adr & ~(BANK1_SIZE - 1);          // Align to Size Boundary

    return (0);
}
#endif



/*
 *  De-Initialize Flash Programming Functions
 *    Parameter:      fnc:  Function Code (1 - Erase, 2 - Program, 3 - Verify)
 *    Return Value:   0 - OK,  1 - Failed
 */

int UnInit (unsigned long fnc) 
{
    (void)fnc;
    
    // Lock Flash
    FLASH->CR  |=  FLASH_LOCK;
    
    return (0);
}

/*
 *  Erase complete Flash Memory
 *    Return Value:   0 - OK,  1 - Failed
 */

int EraseChip (void) 
{
    unsigned long temp;

    temp  = FLASH->CR;
    temp &= ~FLASH_PROG_MASK;     /* Clear all flash operation flag */
    temp |= FLASH_MER;            /* Set Mass Erase flag */

    FLASH->CR  =  temp;                     // Mass Erase Enabled

    FLASH->CR  |=  FLASH_STRT;                    // Start Erase

    while (FLASH->SR  & FLASH_BSY) 
    {
        IWDG->KR = 0xAAAA;                          // Reload IWDG
    }

    FLASH->CR  &= ~FLASH_MER;                     // Mass Erase Disabled

    return (0);                                   // Done
}

/*
 *  Erase Sector in Flash Memory
 *    Parameter:      adr:  Sector Address
 *    Return Value:   0 - OK,  1 - Failed
 */

int EraseSector (unsigned long adr) 
{
    unsigned long temp;
    
    temp  = FLASH->CR;
    temp &= ~FLASH_PROG_MASK;        /* Clear all flash operation flag */
    temp |= FLASH_PER;              /* Set Page Erase flag */

    FLASH->CR  =  temp;                     // Page Erase Enabled

    FLASH->AR   =  adr;                         // Page Address
    FLASH->CR  |=  FLASH_STRT;                  // Start Erase

    while (FLASH->SR  & FLASH_BSY) 
    {
        IWDG->KR = 0xAAAA;                        // Reload IWDG
    }

    FLASH->CR  &= ~FLASH_PER;                   // Page Erase Disabled 

  return (0);                                   // Done
}

/*
 *  Program Page in Flash Memory
 *    Parameter:      adr:  Page Start Address
 *                    sz:   Page Size
 *                    buf:  Page Data
 *    Return Value:   0 - OK,  1 - Failed
 */
int ProgramPage (unsigned long adr, unsigned long sz, unsigned char *buf) 
{
    IWDG->KR = 0xAAAA;
    
    FLASH->CR &= ~FLASH_PROG_MASK;      /* Clear all flash operation flag */

    sz = (sz + 3) & (~3);                           // Adjust size for Words
    
    while (sz) 
    {
      FLASH->CR  |=  FLASH_PG;                  // Programming Enabled

      M32(adr) = *((unsigned int *)buf);      // Program Word
      while (FLASH->SR  & FLASH_BSY);

      FLASH->CR  &= ~FLASH_PG;                  // Programming Disabled

      // Check for Errors
      if (FLASH->SR  & (FLASH_PGERR | FLASH_WRPRTERR)) 
      {
        FLASH->SR  |= FLASH_PGERR | FLASH_WRPRTERR;
        return (1);                             // Failed
      }

      // Go to next Word
      adr += 4;
      buf += 4;
      sz  -= 4;
      
      IWDG->KR = 0xAAAA;
    }
    
    return (0);                                   // Done
}
